Castbars = LibStub("AceAddon-3.0"):NewAddon("Castbars");

Castbars.SharedMedia = LibStub("LibSharedMedia-3.0");

function Castbars:FrameTextureRestore(frame)
    local barTexture = self.SharedMedia:Fetch('statusbar', self.db.profile[frame.configName .. "Texture"]);
    local borderTexture = self.SharedMedia:Fetch('border', self.db.profile[frame.configName .. "Border"]);
    if (barTexture) then
        frame.statusBar:SetStatusBarTexture(barTexture);
    end
    if (borderTexture) then
        frame.backdrop:SetBackdrop({edgeFile = borderTexture, edgeSize = 16, tileSize = 16, insets = {bottom = 5, top = 5, right = 5, left = 5}});
        local brightness = self.db.profile[frame.configName .. "BorderBrightness"];
        frame.backdrop:SetBackdropBorderColor(brightness, brightness, brightness, 1.0);
    end
end

function Castbars:FrameTextureRestoreAll()
    for i, frame in pairs(self.frames) do
        self:FrameTextureRestore(frame)
    end
end

function Castbars:FrameLayoutRestore(frames)
    for i, frame in pairs(frames) do
        -- Position
        local position = self.db.profile[frame.configName .. "Position"];
        if (frame.dragable and position) then
            frame:ClearAllPoints();
            frame:SetPoint(position.point, position.parent, position.relpoint, position.x, position.y);
        end

        -- Texture
        self:FrameTextureRestore(frame)

        -- Dimensions
        frame:SetWidth(self.db.profile[frame.configName .. "Width"]);
        frame:SetHeight(self.db.profile[frame.configName .. "Height"]);

        -- Icon position
        if (frame.icon) then
            frame.icon:ClearAllPoints();
            if (self.db.profile[frame.configName .. "Border"] == "None") then
                frame.icon:SetPoint("RIGHT", frame, "LEFT", 0, 0);
                frame.icon:SetHeight(self.db.profile[frame.configName .. "Height"]);
                frame.icon:SetWidth(self.db.profile[frame.configName .. "Height"]);
            else
                frame.icon:SetPoint("RIGHT", frame, "LEFT", -5, 0);
                frame.icon:SetHeight(self.db.profile[frame.configName .. "Height"] + 5);
                frame.icon:SetWidth(self.db.profile[frame.configName .. "Height"] + 5);
            end
        end

        -- Visibility
        if (self.db.profile[frame.configName .. "Show"] == false) then
            frame.showCastbar = false;
        else
            frame.showCastbar = true;
        end
    end
end

function Castbars:FrameCustomize(frames)
    for i, frame in pairs(frames) do
        local frameName = frame:GetName();

        if (frameName:sub(1, 11) == "MirrorTimer") then
            frame.type = "Mirror";
            frame.index = tonumber(frameName:sub(12,12));
            frame.statusBar = _G[frameName .. "StatusBar"];
            frame.configName = "MirrorTimer";
            if (frame.index == 1) then
                frame.dragable = true;
            end
        elseif (frameName:sub(1, 10) == "CastingBar") then
            frame.type = "Castbar";
            frame.statusBar = frame;
            frame.configName = frameName;
            frame.dragable = true;
        end

        -- Make dragable
        if (frame.dragable) then
            frame:EnableMouse(true);
            frame:SetMovable(true);
            frame:RegisterForDrag("LeftButton");
            frame:SetScript("OnDragStart", function(frame)
                if (self.ConfigMode) then
                    frame:StartMoving();
                end
            end);
            frame:SetScript("OnDragStop", function(frame)
                frame:StopMovingOrSizing();
                if (not self.db.profile[frame.configName .. "Position"]) then
                    self.db.profile[frame.configName .. "Position"] = {};
                end
                local position = self.db.profile[frame.configName .. "Position"];
                position.point, position.parent, position.relpoint, position.x, position.y = frame:GetPoint();
            end);
        end

        -- Adjust spark position
        frame.spark = _G[frameName .. "Spark"];
        if (frame.spark) then
            local setPoint = frame.spark.SetPoint;
            frame.spark.SetPoint = function(self, point, relativeFrame, realativePoint, x, y)
                setPoint(self, point, relativeFrame, realativePoint, x, 0);
            end
        end

        -- Adjust text position
        frame.text = _G[frameName .. "Text"];
        frame.text:ClearAllPoints();
        frame.text:SetPoint("CENTER", frame.statusBar, "CENTER", 0, 0);

        -- Remove the border
        frame.border = _G[frameName .. "Border"];
        frame.border:SetTexture();

        -- Remove the flash
        frame.flash = _G[frameName .. "Flash"];
        if (frame.flash) then
            frame.flash:SetTexture();
        end

        -- Create backdrop
        frame.backdrop = CreateFrame("Frame", frameName .. "Backdrop", frame);
        if (frame.type == "Castbar") then
            frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 0);
        elseif (frame.type == "Mirror") then
            frame.backdrop:SetPoint("CENTER", frame, "CENTER", 0, 3);
        end

        -- Make icon visible
        frame.icon = _G[frameName .. "Icon"];
        if (frame.icon) then
            frame.icon:Show();
        end

        if (frame.type == "Castbar") then
            -- Add timer text
            frame.timer = frame:CreateFontString(nil, "ARTWORK", "GameFontHighlight");
            frame.timer:SetPoint("RIGHT", frame, "RIGHT", -5, 0);
            frame.nextupdate = 0.1;

            -- Add latency indication texture
            frame.latency = frame:CreateTexture(nil, "BACKGROUND");
            frame.latency:SetHeight(frame:GetHeight());
            frame.latency:SetPoint("RIGHT", frame, "RIGHT", 0, 0);
            frame.latency:SetTexture(1, 0, 0, 1);
        end

        if (frame.type == "Mirror") then
            for i, region in pairs({frame:GetRegions()}) do
                if (region.GetTexture and region:GetTexture() == "SolidTexture") then
                    frame.shade = region;
                end
            end
        end

        -- Automatically adjust the height of sub elements
        local setHeight = frame.SetHeight;
        if (frame.type == "Castbar") then
            frame.SetHeight = function(self, height)
                frame.backdrop:SetHeight(height + 10);
                frame.spark:SetHeight(2.5 * height);
                frame.latency:SetHeight(height);
                setHeight(self, height);
            end
        elseif (frame.type == "Mirror") then
            frame.SetHeight = function(self, height)
                frame.backdrop:SetHeight(height + 10);
                frame.statusBar:SetHeight(height);
                frame.shade:SetHeight(height);
                setHeight(self, height + 10);
            end
        end

        -- Automatically adjust the width of sub elements
        local setWidth = frame.SetWidth
        if (frame.type == "Castbar") then
            frame.SetWidth = function(self, width)
                frame.backdrop:SetWidth(width + 10);
                setWidth(self, width);
            end
        elseif (frame.type == "Mirror") then
            frame.SetWidth = function(self, width)
                frame.backdrop:SetWidth(width + 10);
                frame.statusBar:SetWidth(width);
                frame.shade:SetWidth(width);
                setWidth(self, width + 10);
            end
        end

        frame:Hide();
    end
end

function Castbars:GetOptionsTableForBar(frameConfigName, friendlyName, order, enableDisable)
    local optionstable =
    {
        type = "group",
        name = friendlyName .. " Options",
        guiInline = true,
        order = order,
        args =
        {
            width =
            {
                name = "Width",
                type = "range",
                order = 2,
                desc = "Set the width of the " .. friendlyName,
                min = 100,
                max = 300,
                get = function() return self.db.profile[frameConfigName .. "Width"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName .. "Width"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            height =
            {
                name = "Height",
                type = "range",
                order = 3,
                desc = "Set the height of the " .. friendlyName,
                min = 10,
                max = 50,
                get = function() return self.db.profile[frameConfigName .. "Height"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName .. "Height"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            texture =
            {
                name = "Texture",
                type = 'select',
                order = 4,
                desc = "Select texture to use for the " .. friendlyName,
                dialogControl = 'LSM30_Statusbar',
                values = AceGUIWidgetLSMlists.statusbar,
                get = function() return self.db.profile[frameConfigName .. "Texture"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName .. "Texture"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            border =
            {
                name = "Border",
                type = 'select',
                order = 5,
                desc = "Select border to use for the " .. friendlyName,
                dialogControl = 'LSM30_Border',
                values = AceGUIWidgetLSMlists.border,
                get = function() return self.db.profile[frameConfigName .. "Border"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName .. "Border"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
            borderbrightness =
            {
                name = "Border Brightness",
                type = "range",
                order = 6,
                desc = "Set the brightness of the border of the " .. friendlyName,
                min = 0.0,
                max = 1.0,
                isPercent = true,
                get = function() return self.db.profile[frameConfigName .. "BorderBrightness"] end,
                set = function(info, value)
                    self.db.profile[frameConfigName .. "BorderBrightness"] = value;
                    self:FrameLayoutRestore(self.frames);
                end,
            },
        },
    };
    if (enableDisable) then
        optionstable.args.enable =
        {
            name = "Enable",
            type = "toggle",
            order = 1,
            desc = "Toggles display of the " .. friendlyName,
            get = function() return self.db.profile[frameConfigName .. "Show"] end,
            set = function()
                self.db.profile[frameConfigName .. "Show"] = not self.db.profile[frameConfigName .. "Show"];
                self:FrameLayoutRestore(self.frames);
            end,
        }
    end
    return optionstable;
end

function Castbars:GetOptionsTable()
    local optionstable =
    {
        type = "group",
        icon = "",
        name = "Castbars",
        handler = self,
        childGroups = "tree",
        args =
        {
            configmode =
            {
                type = "group",
                name = "Configuration Mode",
                guiInline = true,
                order = 0,
                args =
                {
                    toggleconfigmode =
                    {
                        name = "Toggle",
                        type = "execute",
                        order = 1,
                        desc = "Toggle configuration mode to allow moving bars and setting appearance options",
                        func = "Toggle",
                    },
                },
            },
            player = self:GetOptionsTableForBar("CastingBarFrame", "Player Castbar", 1, true),
            target = self:GetOptionsTableForBar("CastingBarTargetFrame", "Target Castbar", 2, true),
            mirror = self:GetOptionsTableForBar("MirrorTimer", "Mirror Timer Bars", 3),
        },
    };
    return optionstable;
end

function Castbars:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("CastbarsDB",
    {
        profile =
        {
            CastingBarFrameShow = true,
            CastingBarFrameWidth = 195,
            CastingBarFrameHeight = 13,
            CastingBarFrameTexture = "Blizzard",
            CastingBarFrameBorder = "Blizzard Dialog",
            CastingBarFrameBorderBrightness = 0.5,
            CastingBarFramePosition = {point = "BOTTOM", relpoint = "BOTTOM", x = 0, y = 200},

            CastingBarTargetFrameShow = true,
            CastingBarTargetFrameWidth = 195,
            CastingBarTargetFrameHeight = 13,
            CastingBarTargetFrameTexture = "Blizzard",
            CastingBarTargetFrameBorder = "None",
            CastingBarTargetFrameBorderBrightness = 0.5,
            CastingBarTargetFramePosition = {point = "CENTER", relpoint = "CENTER", x = 0, y = 200},

            MirrorTimerWidth = 195,
            MirrorTimerHeight = 13,
            MirrorTimerTexture = "Blizzard",
            MirrorTimerBorder = "Blizzard Dialog",
            MirrorTimerBorderBrightness = 0.5,
            MirrorTimerFramePosition = nil,
        }
    });

    LibStub("AceConfig-3.0"):RegisterOptionsTable("Castbars", self:GetOptionsTable());
    self.BlizzardOptionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Castbars", "Castbars");

    -- Prevent the UIParent from moving the CastingBarFrame around
    UIPARENT_MANAGED_FRAME_POSITIONS["CastingBarFrame"] = nil;

    -- Create target casting bar
    CreateFrame("StatusBar", "CastingBarTargetFrame", UIParent, "CastingBarFrameTemplate");
    CastingBarTargetFrame:SetScript("OnEvent", function(frame, event, ...)
        if (event == "PLAYER_TARGET_CHANGED") then
            event = "PLAYER_ENTERING_WORLD"
        end
        CastingBarFrame_OnEvent(frame, event, ...);
    end)
    CastingBarTargetFrame:RegisterEvent("PLAYER_TARGET_CHANGED");
    CastingBarFrame_OnLoad(CastingBarTargetFrame, "target", false);

    self.frames = {CastingBarFrame, CastingBarTargetFrame, MirrorTimer1, MirrorTimer2, MirrorTimer3};

    -- Customize the bars
    self:FrameCustomize(self.frames);

    -- Restore Frame Layout
    self:FrameLayoutRestore(self.frames);

    -- Listen to LibSharedMedia-3.0 callbacks
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_Registered", "FrameTextureRestoreAll")
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_SetGlobal", "FrameTextureRestoreAll")

    -- Add timer handling code
    hooksecurefunc("CastingBarFrame_OnUpdate", function(self, elapsed)
        if (self.timer) then
            if (self.nextupdate < elapsed) then
                if (self.casting) then
                    self.timer:SetText(format("%.1f", max(self.maxValue - self.value, 0)));
                elseif (self.channeling) then
                    self.timer:SetText(format("%.1f", max(self.value, 0)));
                else
                    self.timer:SetText("");
                end;
                self.nextupdate = 0.1;
            else
                self.nextupdate = self.nextupdate - elapsed;
            end;
        end;
    end);

    -- Add latency indication handling code
    hooksecurefunc("CastingBarFrame_OnShow", function(self)
        if (self.latency) then
            local min, max = self:GetMinMaxValues();
            local down, up, lag = GetNetStats();
            local latency = (lag / 1000) / (max - min);
            if (latency < 0) then latency = 0 elseif (latency > 1) then latency = 1 end
            if (self.channeling) then
                self.latency:ClearAllPoints();
                self.latency:SetPoint("LEFT", self, "LEFT", 0, 0);
            else
                self.latency:ClearAllPoints();
                self.latency:SetPoint("RIGHT", self, "RIGHT", 0, 0);
            end;
            if (max > 4) then
                self.latency:Hide();
            else
                self.latency:SetWidth(self:GetWidth() * latency);
                self.latency:Show();
            end
        end
    end);

    self.GetOptionsTableForBar = nil;
    self.GetOptionsTable = nil;
    self.FrameCustomize = nil;
    self.OnInitialize = nil;
end

--[[ ConfigMode Support ]]--

Castbars.CastingBarFrame_OnEvent = CastingBarFrame_OnEvent;
Castbars.MirrorTimerFrame_OnEvent = MirrorTimerFrame_OnEvent;
Castbars.MirrorTimer_Show = MirrorTimer_Show;
Castbars.DoNothing = function() end;

function Castbars:Show()
    self.ConfigMode = true;
    CastingBarFrame_OnEvent = self.DoNothing;
    MirrorTimerFrame_OnEvent = self.DoNothing
    MirrorTimer_Show = self.DoNothing;

    CastingBarFrameText:SetText("Player Castbar")
    CastingBarFrame:SetStatusBarColor(0.0, 1.0, 0.0);
    CastingBarFrame:SetAlpha(1)
    CastingBarFrameSpark:Hide();
    CastingBarFrameFlash:Hide();
    CastingBarFrame.fadeOut = nil;
    CastingBarFrame:Show();

    CastingBarTargetFrameText:SetText("Target Castbar")
    CastingBarTargetFrame:SetStatusBarColor(0.0, 1.0, 0.0);
    CastingBarTargetFrame:SetAlpha(1)
    CastingBarTargetFrameSpark:Hide();
    CastingBarTargetFrameFlash:Hide();
    CastingBarTargetFrame.fadeOut = nil;
    CastingBarTargetFrame:Show();

    MirrorTimer1.paused = 1;
    MirrorTimer2.paused = 1;
    MirrorTimer3.paused = 1;
    MirrorTimer1Text:SetText("Mirror Timer 1");
    MirrorTimer2Text:SetText("Mirror Timer 2");
    MirrorTimer3Text:SetText("Mirror Timer 3");
    MirrorTimer1StatusBar:SetStatusBarColor(0.0, 1.0, 0.0);
    MirrorTimer2StatusBar:SetStatusBarColor(0.0, 1.0, 0.0);
    MirrorTimer3StatusBar:SetStatusBarColor(0.0, 1.0, 0.0);
    MirrorTimer1StatusBar:SetValue(select(2, MirrorTimer1StatusBar:GetMinMaxValues()));
    MirrorTimer2StatusBar:SetValue(select(2, MirrorTimer2StatusBar:GetMinMaxValues()));
    MirrorTimer3StatusBar:SetValue(select(2, MirrorTimer3StatusBar:GetMinMaxValues()));
    MirrorTimer1:Show();
    MirrorTimer2:Show();
    MirrorTimer3:Show();
end

function Castbars:Hide()
    MirrorTimer1:Hide();
    MirrorTimer2:Hide();
    MirrorTimer3:Hide();
    CastingBarFrame:Hide();
    CastingBarTargetFrame:Hide();
    CastingBarFrame_OnEvent = self.CastingBarFrame_OnEvent;
    MirrorTimerFrame_OnEvent = self.MirrorTimerFrame_OnEvent;
    MirrorTimer_Show = self.MirrorTimer_Show;
    self.ConfigMode = false;
end

function Castbars:Toggle()
    if (self.ConfigMode) then
        self:Hide();
    else
        self:Show();
    end
end

CONFIGMODE_CALLBACKS = CONFIGMODE_CALLBACKS or {};
CONFIGMODE_CALLBACKS["Castbars"] = function(action)
    if (action == "ON") then Castbars:Show()
    elseif (action == "OFF") then Castbars:Hide() end
end
